import pytest
import os
import joblib
from typing import Union

import numpy as np
import pandas as pd

from sklearn.datasets import make_classification, make_regression
from sklearn.base import clone
from sklearn.metrics import SCORERS

from Amplo.Classifiers import CatBoostClassifier, LGBMClassifier, XGBClassifier, StackingClassifier
from Amplo.Regressors import CatBoostRegressor, LGBMRegressor, XGBRegressor, StackingRegressor


Model = Union[CatBoostClassifier, LGBMClassifier, XGBClassifier, StackingClassifier,
              CatBoostRegressor, LGBMRegressor, XGBRegressor, StackingRegressor]


setup_class_params = [CatBoostClassifier, LGBMClassifier, XGBClassifier, StackingClassifier,
                      CatBoostRegressor, LGBMRegressor, XGBRegressor, StackingRegressor]


@pytest.fixture(scope='class', params=setup_class_params)
def setup_class(request):
    model = request.param

    if 'Classifier' in model.__name__:
        x, y = make_classification(n_classes=5, n_informative=15)
        is_classification = True
    elif 'Regressor' in model.__name__:
        x, y = make_regression(n_informative=15)
        is_classification = False
    else:
        raise ValueError('Invalid model requested')

    if 'Stacking' in model.__name__:
        if is_classification:
            model_params = {'CatBoostClassifier': {'depth': 10}}
        else:
            model_params = {'CatBoostRegressor': {'depth': 10}}
    else:
        model_params = {'max_depth': 10}

    request.cls._model = model
    request.cls.model_params = model_params
    request.cls.is_classification = is_classification
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.mark.usefixtures('setup_class')
class TestClassifier:

    @pytest.fixture(autouse=True)
    def setup(self):
        # Initialize model
        self.model = self._model()
        yield

    def test_set_params(self):
        self.model.set_params(**self.model_params)

    def test_get_params(self):
        self.model.get_params()

    def test_fit_pandas(self):
        self.model.fit(self.x, self.y)

    def test_fit_numpy(self):
        self.model.fit(self.x.to_numpy(), self.y.to_numpy())

    def test_trained_attr(self):
        assert hasattr(self.model, 'trained')
        assert self.model.trained is False
        self.model.fit(self.x, self.y)
        assert self.model.trained is True

    def test_predict(self):
        self.model.fit(self.x, self.y)
        prediction = self.model.predict(self.x)

        assert len(prediction.shape) == 1
        if self.is_classification:
            assert np.allclose(prediction.astype('int'), prediction)

    def test_predict_proba(self):
        if self.is_classification:
            self.model.fit(self.x, self.y)
            prediction = self.model.predict_proba(self.x)

            assert not np.isnan(prediction).any(), 'NaN in prediction: {}'.format(prediction)
            assert len(prediction.shape) == 2
            assert prediction.shape[1] == len(np.unique(self.y))
            assert np.allclose(np.sum(prediction, axis=1), 1)

    def test_cloneable(self):
        clone_model: Model = clone(self.model.set_params(**self.model_params))  # noqa
        clone_model.fit(self.x, self.y)

        if self.is_classification:
            prediction = clone_model.predict_proba(self.x)
            assert len(prediction.shape) != 1
            assert np.allclose(np.sum(prediction, axis=1), 1)
        else:
            prediction = clone_model.predict(self.x)
            assert len(prediction.shape) == 1

    def test_scorer(self):
        self.model.fit(self.x, self.y)

        if self.is_classification:
            scorers = ['neg_log_loss', 'accuracy', 'f1_micro']
        else:
            scorers = ['neg_mean_squared_error', 'neg_mean_absolute_error', 'r2']

        for name in scorers:
            SCORERS[name](self.model, self.x, self.y)

    def test_binary(self):
        if self.is_classification:
            x, y = make_classification()
            self.model.fit(x, y)
            self.model.predict(x)
            self.model.predict_proba(x)
            SCORERS['neg_log_loss'](self.model, x, y)
            SCORERS['accuracy'](self.model, x, y)
            SCORERS['f1_micro'](self.model, x, y)

    def test_pickleable(self):
        x, y = make_classification()
        self.model.fit(x, y)
        joblib.dump(self.model, 'tmp.joblib')
        os.remove('tmp.joblib')



import pytest

import json
import numpy as np
import pandas as pd
from sklearn.datasets import load_iris, fetch_california_housing

# TODO: Make use of the dummy data creator
#  from Amplo.Utils.testing import (DummyDataSampler, make_data, make_cat_data, make_num_data)
from Amplo.AutoML import DataProcessor
from Amplo.Utils.data import check_dataframe_quality


@pytest.fixture(scope='class', params=['regression', 'classification'])
def make_mode(request):
    mode = request.param
    target = 'target'
    if mode == 'classification':
        x, y = load_iris(return_X_y=True, as_frame=True)
    elif mode == 'regression':
        x, y = fetch_california_housing(return_X_y=True, as_frame=True)
    else:
        raise NotImplementedError('Invalid mode')
    x.columns = [f'Feature_{i}' for i in range(len(x.columns))]
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = pd.concat([x, y.to_frame(target)], axis=1)
    yield


@pytest.mark.usefixtures('make_mode')
class TestMode:
    def test_mode(self):
        dp = DataProcessor(self.target)
        cleaned = dp.fit_transform(self.data)
        assert check_dataframe_quality(cleaned)


class TestDataProcessor:

    def test_interpolation(self):
        dp = DataProcessor()
        data = pd.DataFrame({'a': [1, np.nan, np.nan, np.nan, 5], 'b': [1, 2, 3, 4, 5]})
        cleaned = dp.fit_transform(data)
        assert cleaned['a'].tolist() == [1, 0, 0, 0, 5]

    def test_type_detector(self):
        dp = DataProcessor()
        data = pd.DataFrame({
            'a': ['a', 'b', 'c', 'd', 'a'],
            'b': [1, 2, 3, 4, 5],
            'c': ['2020-01-01', '2020-01-02', '2020-01-03', '2020-01-04', '2020-01-05']})
        cleaned = dp.fit_transform(data)
        assert pd.api.types.is_integer_dtype(cleaned['b'])
        assert pd.api.types.is_datetime64_any_dtype(cleaned['c'])
        assert 'a_b' in cleaned
        assert 'a_c' in cleaned
        assert 'a_d' in cleaned

    def test_type_detector_with_nan(self):
        dp = DataProcessor()
        data = pd.DataFrame({
            'a': ['a', 'b', np.nan, 'c', 'd'],
            'b': [1, 2, 3, 4, np.nan],
            'c': ['2020-01-01', np.nan, '2020-01-03', '2020-01-04', '2020-01-05']})
        cleaned = dp.fit_transform(data)
        assert pd.api.types.is_float_dtype(cleaned['b'])
        assert pd.api.types.is_datetime64_any_dtype(cleaned['c'])
        assert 'a_b' in cleaned
        assert 'a_c' in cleaned
        assert 'a_d' in cleaned

    def test_missing_values(self):
        data = pd.DataFrame({
            'a': ['a', 'b', np.nan, 'c', 'd'],
            'b': [1, 2, 3, 4, np.nan],
            'c': ['2020-01-01', np.nan, '2020-01-03', '2020-01-04', '2020-01-05'],
            'd': [1, 2, 3, 4, 5]})

        # Remove rows
        dp = DataProcessor(missing_values='remove_rows')
        cleaned = dp.fit_transform(data)
        assert len(cleaned) == 3, cleaned.head()        # the nan in the categorical is omitted

        # Remove cols
        dp = DataProcessor(missing_values='remove_cols')
        cleaned = dp.fit_transform(data)
        assert len(cleaned.keys()) == 6, cleaned.head() # Categorical NaN is allowed, so 5 categoricals and d

        # Replace with 0
        dp = DataProcessor(missing_values='zero')
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ['a_b', 'a_c', 'a_d']] == 0).all()
        assert cleaned['b'][4] == 0
        assert cleaned['c'][1] == 0

        # Interpolate
        dp = DataProcessor(missing_values='interpolate')
        cleaned = dp.fit_transform(data)
        assert cleaned.isna().sum().sum() == 0, cleaned.head()

        # Fill with mean
        dp = DataProcessor(missing_values='mean')
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ['a_b', 'a_c', 'a_d']] == 0).all()
        assert cleaned['b'][4] == 2.5
        assert cleaned['c'][1] == pd.to_datetime('2020-01-03 6:00:00', utc=True)

    def test_classification_target(self):
        data = pd.DataFrame({'a': [2, 2, 1, 1, 2], 'b': ['class1', 'class2', 'class1', 'class2', 'class1']})

        # Numerical not starting at 0
        dp = DataProcessor(target='a')
        cleaned = dp.fit_transform(data)
        assert set(cleaned['a']) == {0, 1}

        # Categorical
        dp = DataProcessor(target='b')
        cleaned = dp.fit_transform(data)
        assert set(cleaned['b']) == {0, 1}

    def test_outliers(self):
        x = pd.DataFrame({'a': [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1e15],
                          'target': np.linspace(0, 1, 24).tolist()})

        # Clip
        dp = DataProcessor(outlier_removal='clip', target='target')
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"

        # z-score
        dp = DataProcessor(outlier_removal='z-score', target='target')
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"
        assert np.isclose(dp.transform(pd.DataFrame({'a': [1e14], 'b': [1]})).max().max(), 1e14)
        assert dp.transform(pd.DataFrame({'a': [1e16], 'b': [1]})).max().max() == 1

        # Quantiles
        dp = DataProcessor(outlier_removal='quantiles', target='target')
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"
        assert dp.transform(pd.DataFrame({'a': [2], 'b': [-2]})).max().max() == 0

    def test_duplicates(self):
        # Create a DataFrame that contains two columns with the same name
        x = pd.DataFrame({'a': [1, 2, 1], 'a_copy': [1, 2, 1], 'b': [3, 1, 3]})
        x = x.rename(columns={'a_copy': 'a'})
        dp = DataProcessor()
        xt = dp.fit_transform(x)
        assert len(xt) == 2, "Didn't remove duplicate rows"
        assert len(xt.keys()) == 2, "Didn't remove duplicate columns"

    def test_constants(self):
        x = pd.DataFrame({'a': [1, 1, 1, 1, 1], 'b': [1, 2, 3, 5, 6]})
        dp = DataProcessor()
        xt = dp.fit_transform(x)
        assert 'a' not in xt.keys(), "Didn't remove constant column"

    def test_dummies(self):
        x = pd.DataFrame({'a': ['a', 'b', 'c', 'b', 'c', 'a']})
        dp = DataProcessor(cat_cols=['a'])
        xt = dp.fit_transform(x)
        assert 'a' not in xt.keys(), "'a' still in keys"
        assert 'a_b' in xt.keys(), "a_b missing"
        assert 'a_c' in xt.keys(), "a_c missing"
        xt2 = dp.transform(pd.DataFrame({'a': ['a', 'c']}))
        assert np.allclose(xt2.values, pd.DataFrame({'a_a': [1, 0], 'a_b': [0, 0], 'a_c': [0, 1]}).values), \
            "Converted not correct"

    def test_nan_categorical(self):
        # Setup
        df = pd.DataFrame({'a': ['hoi', np.nan, np.nan, np.nan]})
        dp = DataProcessor()
        df = dp.fit_transform(df)

        # Tests
        assert 'a' in dp.cat_cols, 'NaN categorical column not recognised as categorical.'
        assert 'a_hoi' in list(df.keys()), f'Categorical column not properly converted: {df.keys()}'

    def test_settings(self):
        # todo add tests for different arguments, was failing for outlier_removal=z-score.
        target = 'target'
        x = pd.DataFrame({'a': ['a', 'b', 'c', 'b', 'c', 'a'],
                          'b': [1, 1, 1, 1, 1, 1],
                          target: ['a', 'b', 'c', 'b', 'c', 'a']})
        dp = DataProcessor(cat_cols=['a'], target=target)
        xt = dp.fit_transform(x)
        assert len(xt.drop(target, axis=1).keys()) == x['a'].nunique()
        settings = dp.get_settings()
        dp2 = DataProcessor()
        dp2.load_settings(settings)
        assert isinstance(dp2.get_settings()['_label_encodings'], list)
        xt2 = dp2.transform(pd.DataFrame({'a': ['a', 'b'], 'b': [1, 2]}))
        assert np.allclose(pd.DataFrame({'b': [1.0, 2.0], 'a_a': [1, 0], 'a_b': [0, 1], 'a_c': [0, 0]}).values,
                           xt2.values)

    def test_pruner(self):
        x = pd.DataFrame({'a': ['a', 'b', 'c', 'b', 'c', 'a'], 'b': [1, 1, 1, 1, 1, 1]})
        dp = DataProcessor()
        dp.fit_transform(x)
        dp.prune_features(['b'])
        assert dp.int_cols == ['b']
        assert dp.cat_cols == []

    def test_json_serializable(self):
        x = pd.DataFrame({'a': ['a', 'b', 'c', 'b', 'c', 'a'], 'b': [1, 'missing', 1, 1, 1, 1]})
        for o in ['quantiles', 'z-score', 'clip', 'none']:
            for mv in ['remove_rows', 'remove_cols', 'interpolate', 'mean', 'zero']:
                dp = DataProcessor(outlier_removal=o, missing_values=mv)
                dp.fit_transform(x)
                json.dumps(dp.get_settings())

    def test_cat_target(self):
        df = pd.DataFrame({'a': ['a', 'b', 'c', 'b', 'c', 'a'], 'b': [1, 2, 3, 4, 5, 6]})
        dp = DataProcessor(target='a')
        xt = dp.fit_transform(df)
        assert 'a' in xt
        assert np.allclose(xt['a'].values, np.array([0, 1, 2, 1, 2, 0]))

    def test_target_encoding(self):
        target = 'target'
        orig_labels = pd.Series(['a', 'b', 'c', 'b', 'c', 'a'], name=target)
        dp = DataProcessor(target=target)
        enc_labels = pd.Series(dp.encode_labels(orig_labels))
        assert enc_labels.nunique() == orig_labels.nunique()
        assert enc_labels.min() == 0, 'Encoding must start at zero'
        assert pd.api.types.is_integer_dtype(enc_labels), 'Encoding must be of dtype `int`'
        dec_labels = pd.Series(dp.decode_labels(enc_labels))
        assert (dec_labels == orig_labels).all(), 'Decoding does not result in original dataframe'



import shutil
import os

import numpy as np


def rmtree(folder="AutoML", must_exist=False):
    if must_exist and not os.path.exists(folder):
        raise FileNotFoundError(f"Directory {folder} does not exist")
    if os.path.exists(folder):
        shutil.rmtree(folder)


# ----------------------------------------------------------------------
# Dummies


class _RandomClassifier:
    """
    Dummy classifier for testing.
    """

    def __init__(self):
        self.classes = None

    def fit(self, x, y):
        self.classes = np.unique(y)

    def predict(self, x):
        return np.random.choice(self.classes, len(x))

    def predict_proba(self, x):
        size = len(x), len(self.classes)
        proba = np.random.uniform(size=size)
        return proba * (1.0 / proba.sum(1)[:, np.newaxis])  # normalize


class _RandomRegressor:
    """
    Dummy regressor for testing.
    """

    def __init__(self):
        self.range = None

    def fit(self, x, y):
        self.range = np.min(y), np.max(y)

    def predict(self, x):
        return np.random.uniform(*self.range, len(x))


class RandomPredictor:
    """
    Dummy predictor for testing.

    Parameters
    ----------
    mode : str
        Predicting mode ("classification" or "regression").
    """

    def __init__(self, mode):
        if mode == "classification":
            self.predictor = _RandomClassifier()
        elif mode == "regression":
            self.predictor = _RandomRegressor()
        else:
            raise ValueError("Invalid predictor mode.")

    def fit(self, x, y):
        return self.predictor.fit(x, y)

    def predict(self, x):
        return self.predictor.predict(x)

    def predict_proba(self, x):
        assert isinstance(self.predictor, _RandomClassifier)
        return self.predictor.predict_proba(x)

    @property
    def classes_(self):
        if hasattr(self.predictor, "classes"):
            return self.predictor.classes


class OverfitPredictor:
    """
    Dummy predictor for testing.

    Parameters
    ----------
    mode : str
        Predicting mode ("classification" or "regression").
    """

    def __init__(self, mode):
        if mode == "classification":
            self.predictor = _OverfitClassifier()
        elif mode == "regression":
            self.predictor = _OverfitRegressor()
        else:
            raise ValueError("Invalid predictor mode.")

    def fit(self, x, y):
        return self.predictor.fit(x, y)

    def predict(self, x):
        return self.predictor.predict(x)

    def predict_proba(self, x):
        assert isinstance(self.predictor, _OverfitClassifier)
        return self.predictor.predict_proba(x)

    @property
    def classes_(self):
        if hasattr(self.predictor, "classes"):
            return self.predictor.classes


class _OverfitClassifier:
    """
    Dummy classifier for testing. Returns the class if present in the data, else
    predicts 0
    """

    def __init__(self):
        self.classes = None
        self.x = None
        self.y = None

    def fit(self, x, y):
        self.x = x.to_numpy()
        self.y = y
        self.classes = y.unique()

    def predict(self, x):
        yt = []
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(-1)
            else:
                yt.append(self.y.iloc[ind[0]])
        return yt

    def predict_proba(self, x):
        yt = []
        zeroes = [0 for _ in range(len(self.classes))]
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(zeroes)
            else:
                yt.append(
                    [
                        0 if self.y.iloc[ind[0]] != i else 1
                        for i in range(len(self.classes))
                    ]
                )
        return yt


class _OverfitRegressor:
    """
    Dummy regressor for testing.
    """

    def __init__(self):
        self.classes = None
        self.x = None
        self.y = None

    def fit(self, x, y):
        self.x = x.to_numpy()
        self.y = y
        self.classes = y.unique()

    def predict(self, x):
        yt = []
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(-1)
            else:
                yt.append(self.y.iloc[ind[0]])
        return yt



from .Sequencer import Sequencer
from .Modeller import Modeller
from .Documenter import Documenter
from .DataSampler import DataSampler
from .DataExplorer import DataExplorer
from .DataProcessor import DataProcessor
from .DataProcessor import DataProcesser
from .DriftDetector import DriftDetector
from .FeatureProcessor import FeatureProcessor
from .FeatureProcessor import FeatureProcesser
from .IntervalAnalyser import IntervalAnalyser
from .Standardizer import Standardizer



import pytest

import pandas as pd
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import KFold, StratifiedKFold

from Amplo import Pipeline
from Amplo.AutoML import Modeller
from Amplo.GridSearch import BaseGridSearch, HalvingGridSearch, OptunaGridSearch
from Amplo.Utils.io import parse_json


@pytest.fixture(scope='class', params=['regression', 'classification'])
def make_data(request):
    mode = request.param

    if mode == 'regression':
        request.cls.k_fold = KFold
        request.cls.objective = 'r2'
        x, y = make_regression()
    elif mode == 'classification':
        request.cls.k_fold = StratifiedKFold
        request.cls.objective = 'r2'
        x, y = make_classification()
    else:
        raise ValueError('Mode is invalid')

    request.cls.mode = mode
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.mark.usefixtures('make_data')
class TestGridSearch:

    @pytest.mark.parametrize('test_case', ['grid_search_type', 'grid_search_iterations'])
    def test_no_grid_search(self, test_case):
        # Set keyword arguments
        kwargs = dict(extract_features=False, sequence=False, plot_eda=False)
        if test_case == 'grid_search_type':
            kwargs.update({test_case: None})
        elif test_case == 'grid_search_iterations':
            kwargs.update({test_case: 0})
        else:
            raise ValueError('Invalid test case')
        # Fit pipeline
        pipeline = Pipeline(**kwargs)
        pipeline.fit(self.x, self.y)
        # Check results
        assert any(pipeline.results.loc[:, 'type'] == 'Hyper Parameter'), 'No hyper parameter results were found'

    @pytest.mark.parametrize('grid_search', [BaseGridSearch, HalvingGridSearch, OptunaGridSearch])
    @pytest.mark.parametrize('simulate', ['small', 'big'])
    def test_each_model(self, grid_search, simulate):
        """
        Given the models which depend on 1) mode and 2) num_samples,
        each model is trained on the given grid search class.

        Thanks to the fact that all grid search models have the same interface,
        this is possible :-)
        """

        if simulate == 'small':
            n_samples = 100
        elif simulate == 'big':
            n_samples = 50_000
        else:
            raise ValueError('Invalid parameter')

        modeller_kwargs = dict(
            mode=self.mode,
            objective=self.objective,
            samples=n_samples,
        )
        models = Modeller(**modeller_kwargs).return_models()

        for model in models:
            # Grid search
            search = grid_search(model,
                                 cv=self.k_fold(n_splits=3),
                                 verbose=0,
                                 timeout=10,
                                 candidates=2,
                                 scoring=self.objective)
            results = search.fit(self.x, self.y)

            # Tests
            model_name = type(model).__name__
            assert isinstance(results, pd.DataFrame), f'Expected result to be pandas.DataFrame' \
                                                      f'but found {type(results)} in {model_name} instead'
            results_keys = ['worst_case', 'mean_objective', 'std_objective', 'params', 'mean_time', 'std_time']
            assert all(key in results.keys() for key in results_keys), f'Keys are missing in {model_name}'
            assert len(results) > 0, f'Results are empty in {model_name}'
            model.set_params(**parse_json(results.iloc[0]['params']))



import numpy as np
import pytest

from Amplo import Pipeline
from Amplo.Observation._model_observer import ModelObserver
from Amplo.Observation.base import ProductionWarning
from tests import OverfitPredictor, RandomPredictor


@pytest.fixture
def make_one_to_one_data(mode):
    size = 100
    if mode == "classification":
        linear_col = np.random.choice([0, 1, 2], size)
    elif mode == "regression":
        linear_col = np.random.uniform(0.0, 1.0, size)
    else:
        raise ValueError("Invalid mode")
    x = linear_col.reshape(-1, 1)
    y = linear_col.reshape(-1)
    yield x, y


class TestModelObserver:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_better_than_linear(self, mode, make_one_to_one_data):
        x, y = make_one_to_one_data

        # Make pipeline and simulate fit
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline.best_model = RandomPredictor(mode=mode)

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_better_than_linear()

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_noise_invariance(self, mode, make_one_to_one_data):
        x, y = make_one_to_one_data

        # Make pipeline and simulate fit
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline.best_model = OverfitPredictor(mode=mode)

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_noise_invariance()

        # Should not trigger normally
        pipeline.best_model = RandomPredictor(mode=mode)
        obs = ModelObserver(pipeline=pipeline)
        obs.check_noise_invariance()

    def test_slice_invariance(self):
        """
        This is a complex test. Slice invariance will be triggered with a linear
        model, when 90% of the data is linearly separable, but 10% is displaced
        compared to the fit.

        We just do this for classification for ease, the observer runs
        irrespective of mode.
        """
        # Classification dataset
        x = np.linspace(0, 10, 100)
        y = np.concatenate(
            (
                np.zeros(48),
                np.ones(48),
                np.zeros(4),
            )
        )

        # Make pipeline and fit
        pipeline = Pipeline(grid_search_iterations=0, document_results=False)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline._data_processing()
        pipeline._feature_processing()
        pipeline.conclude_fitting(
            model="LogisticRegression", params={}, feature_set="RFI"
        )

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_slice_invariance()

        # Should not trigger normally
        pipeline.best_model = RandomPredictor(mode="classification")
        obs = ModelObserver(pipeline=pipeline)
        obs.check_slice_invariance()



import pytest

import pandas as pd
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression

from tests import rmtree


@pytest.fixture(autouse=True)
def rmtree_automl():
    folder = 'AutoML'
    rmtree(folder, must_exist=False)
    yield folder
    rmtree(folder, must_exist=False)


@pytest.fixture
def make_x_y(mode):
    if mode == 'classification':
        x, y = make_classification()
    elif mode == 'regression':
        x, y = make_regression()
    else:
        raise ValueError('Invalid mode')
    yield pd.DataFrame(x), pd.Series(y)



import json
import re

import numpy as np
import pytest

from Amplo import Pipeline
from Amplo.Observation._data_observer import DataObserver
from Amplo.Observation.base import ProductionWarning


class TestDataObserver:
    def test_monotonic_columns(self):
        size = 100
        monotonic = np.array(range(-10, size - 10)) * 4.2  # start=-10, step=4.2
        random = np.random.normal(size=size)
        x = np.concatenate([monotonic[:, None], random[:, None]], axis=1)
        y = random  # does not matter

        # Observe
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_monotonic_columns()
        msg = str(record[0].message)
        monotonic_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert monotonic_cols == [0], "Wrong monotonic columns identified."

    def test_minority_sensitivity(self):
        # Setup
        x = np.hstack(
            (
                np.random.normal(size=(100, 1)),
                np.concatenate((np.zeros((2, 1)), np.random.normal(100, 5, (98, 1)))),
            )
        )
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Observe
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_minority_sensitivity()
        msg = str(record[0].message)
        sensitive_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert sensitive_cols == [1], "Wrong minority sensitive columns identified."

    def test_categorical_mismatch(self):
        # Setup
        x = np.hstack(
            (
                np.array(["New York"] * 50 + ["new-york"] * 50).reshape((-1, 1)),
                np.random.normal(100, 5, (100, 1)),
            )
        )
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Observe
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_categorical_mismatch()
        msg = str(record[0].message)
        sensitive_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert sensitive_cols == [
            {"0": ["new-york", "New York"]}
        ], "Wrong categorical mismatch columns identified."

    def test_extreme_values(self):
        # Setup
        x = np.vstack((np.random.normal(size=100), np.linspace(1000, 10000, 100))).T
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Observe
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_extreme_values()
        msg = str(record[0].message)
        extreme_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert extreme_cols == [1], "Wrong minority sensitive columns identified."



import copy
from datetime import datetime
import re
import time
from typing import Dict, Union
import warnings

import numpy as np
import optuna
import pandas as pd

from Amplo.GridSearch._GridSearch import _GridSearch

__all__ = ['OptunaGridSearch']


class OptunaGridSearch(_GridSearch):

    def __init__(self, model, *args, **kwargs):
        """
        Wrapper for ``optuna`` grid search.

        Takes any model supported by `Amplo.AutoML.Modelling` whose parameter
        search space is predefined for each model.

        Parameters
        ----------
        model : Amplo.AutoML.Modeller.ModelType
            Model object to optimize.
        args
            Arguments passed to `_GridSearch`.
        kwargs
            Keyword arguments passed to `_GridSearch`.
        """
        super().__init__(model, *args, **kwargs)
        if 'params' in kwargs:
            warnings.warn('Parameter `params` has no effect')

        # Counter to keep track of number of trials
        self.trial_count = -1

    @staticmethod
    def _get_suggestion(trial, p_name, p_value):
        """
        Get suggestion for next hyperparameter.

        Parameters
        ----------
        trial : optuna.Trial
            The trial to sample a suggestion from it.
        p_name : str
            Name of the hyperparameter.
        p_value : (str, list) or (str, list, int)
            Sampling information for hyperparameter (dtype, range).

        Returns
        -------
        None or bool or int or float or str
            A specific parameter for given hyperparameter.
        """

        # Read out
        p_type = p_value[0]  # parameter type (str)
        p_args = p_value[1]  # parameter arguments (list, tuple)
        p_kwargs = {}  # additional parameter keyword arguments, depends on parameter type

        # Sanity checks
        assert len(p_args) == 2 or p_type == 'categorical', \
            'Only categorical parameter can have more/less than two suggest args'

        # Find trial.suggest_{...} type
        if p_type == 'categorical':
            p_args = [p_args]  # enclose since will be expanded afterwards
            suggest = trial.suggest_categorical
        elif p_type == 'int':
            suggest = trial.suggest_int
        elif p_type == 'logint':
            p_kwargs += dict(log=True)
            suggest = trial.suggest_int
        elif p_type == 'uniform':
            suggest = trial.suggest_uniform
        elif p_type == 'loguniform':
            suggest = trial.suggest_loguniform
        else:
            raise NotImplementedError('Invalid parameter specification')

        # Suggest parameter given the arguments
        return suggest(p_name, *p_args)

    def _get_hyper_params(self, trial) -> Dict[str, Union[None, bool, int, float, str]]:
        """Use trial to sample from available grid search parameters

        Parameters
        ----------
        trial : optuna.Trial
            The trial from optuna study.

        Returns
        -------
        params : dict of {str : None or bool or int or float or str}
            Sampled grid search parameters.
        """

        param_values = self._hyper_parameter_values
        conditionals = param_values.pop('CONDITIONALS', {})
        params = {}

        # Add non-conditional parameter suggestions (do not depend on presence of another)
        for name, value in param_values.items():
            params[name] = self._get_suggestion(trial, name, value)

        # Add conditional parameter suggestions that satisfy condition:
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                # Add parameters only if criterion is satisfied
                if matching_value != params.get(check_p_name, None):
                    # No match
                    continue
                for name, value in additional_params.items():
                    params[name] = self._get_suggestion(trial, name, value)

        return params

    def fit(self, x, y):
        if isinstance(y, pd.DataFrame):
            assert len(y.keys()) == 1, 'Multiple target columns not supported.'
            y = y[y.keys()[0]]
        assert isinstance(x, pd.DataFrame), 'X should be Pandas DataFrame'
        assert isinstance(y, pd.Series), 'Y should be Pandas Series or DataFrame'

        # Set mode
        self.binary = y.nunique() == 2
        self.samples = len(y)

        # Store
        self.x, self.y = x, y

        # Set up study
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=236868),
            direction='maximize', pruner=_BadTrialPruner(2.0, 15))
        study.optimize(
            self.objective, timeout=self.timeout, n_trials=self.nTrials,
            callbacks=[_StopStudyWhenConsecutivePruning(10), _StopStudyAfterNPruning(50)])

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame({
            'date': datetime.today().strftime('%d %b %y'),
            'model': type(self.model).__name__,
            'params': [x.params for x in study.get_trials()],
            'mean_objective': optuna_results['value'],
            'std_objective': optuna_results['user_attrs_std_value'],
            'worst_case': optuna_results['value'] - optuna_results['user_attrs_std_value'],
            'mean_time': optuna_results['user_attrs_mean_time'],
            'std_time': optuna_results['user_attrs_std_time']
        })
        self.trial_count = len(study.trials)

        return results

    def objective(self, trial: optuna.Trial):
        # Metrics
        scores = []
        times = []
        master = copy.deepcopy(self.model)

        # Cross validation
        for step, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            # Split data
            xt, yt = self.x.iloc[t], self.y.iloc[t]
            xv, yv = self.x.iloc[v], self.y.iloc[v]

            # Train model
            t_start = time.time()
            model = copy.deepcopy(master)
            model.set_params(**self._get_hyper_params(trial))
            model.fit(xt, yt)

            # Results
            score = self.scoring(model, xv, yv)
            scores.append(score)
            times.append(time.time() - t_start)

            # Report intermediate values
            trial.report(score, step)

            # Stop cross validation when it is not promising
            if trial.should_prune():
                break

        # Set manual metrics
        mean_score = np.mean(scores)
        trial.set_user_attr('mean_time', np.mean(times))
        trial.set_user_attr('std_time', np.std(times))
        trial.set_user_attr('mean_value', np.mean(scores))
        trial.set_user_attr('std_value', np.std(scores))

        # Stop trial (avoid overwriting)
        if trial.number == self.nTrials:
            trial.study.stop()

        # Pruning
        if trial.should_prune():
            raise optuna.exceptions.TrialPruned

        return mean_score


# ----------------------------------------------------------------------
# Pruning

class _BadTrialPruner(optuna.pruners.BasePruner):
    """
    Pruner to detect outlying metrics of the trials.

    Prune if the mean intermediate value is worse than the best trial value minus (or plus) a
    multiple of its intermediate value standard deviation.

    Parameters
    ----------
    std_threshold_multiplier : float
        Multiplier for the best trials intermediate value std to define the pruning threshold.
    n_startup_trials : int
        Pruning is disabled until the given number of trials finish in the same study.
    """

    def __init__(self, std_threshold_multiplier, n_startup_trials):
        if n_startup_trials < 0:
            raise ValueError(
                f'Number of startup trials cannot be negative but got {n_startup_trials}.')

        self._std_threshold_multiplier = std_threshold_multiplier
        self._n_startup_trials = n_startup_trials

    def prune(self, study, trial):
        # Don't prune while startup trials
        all_trials = study.get_trials(deepcopy=False)
        n_trials = len([t for t in all_trials if t.state == optuna.trial.TrialState.COMPLETE])
        if n_trials < self._n_startup_trials:
            return False

        # Define pruning thresholds
        best_trial_value = study.best_trial.value
        best_trial_value_std = np.std(list(
            study.best_trial.intermediate_values.values()))
        threshold = best_trial_value - self._std_threshold_multiplier * best_trial_value_std

        # Pruning
        curr_trial_mean = np.mean(list(trial.intermediate_values.values()))
        if study.direction == optuna.study.StudyDirection.MAXIMIZE:
            return curr_trial_mean < threshold
        else:
            raise RuntimeError(
                'Pruning with a threshold is arbitrary when the study direction is undefined.')


class _StopStudyWhenConsecutivePruning:
    """
    Optuna study callback that stops the study when trials keep being pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for consecutively pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(
                f'Threshold cannot be negative but got {threshold}.')

        self._threshold = threshold
        self._consecutive_pruned_count = 0

    def __call__(self, study, trial):
        """
        Stops study when consecutive prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this object.
        """
        if trial.state == optuna.trial.TrialState.PRUNED:
            self._consecutive_pruned_count += 1
        else:
            self._consecutive_pruned_count = 0

        if self._consecutive_pruned_count > self._threshold:
            study.stop()


class _StopStudyAfterNPruning:
    """
    Optuna study callback that stops the study after begin N times pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for total number of pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(
                f'Threshold cannot be negative but got {threshold}.')

        self._threshold = threshold

    def __call__(self, study, trial):
        """
        Stops study when total prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this object.
        """
        all_trials = study.get_trials(deepcopy=False)
        n_prunings = len([t for t in all_trials if t.state == optuna.trial.TrialState.PRUNED])

        if n_prunings > self._threshold:
            study.stop()



import re
import logging
import warnings


__all__ = ["get_model", "hist_search"]


def get_model(model_str, **kwargs):
    # Import here to prevent ImportError (due to circular import)
    from Amplo.AutoML.Modeller import Modeller

    try:
        model = [
            mod
            for mod in Modeller(**kwargs).return_models()
            if type(mod).__name__ == model_str
        ]
        return model[0]
    except IndexError:
        raise IndexError("Model not found.")


def hist_search(array, value):
    """
    Binary search that finds the index in order to fulfill
    ``array[index] <= value < array[index + 1]``

    Parameters
    ----------
    array : array of float
    value : float

    Returns
    -------
    int
        Bin index of the value
    """

    # Return -1 when no bin exists
    if value < array[0] or value >= array[-1]:
        logging.debug(
            f"No bin (index) found for value {value}. Array(Min: {array[0]}, "
            "Max: {array[-1]})"
        )
        return -1

    # Initialize min and max bin index
    low = 0
    high = len(array) - 1

    # Bin search
    countdown = 30
    while countdown > 0:
        # Count down
        countdown -= 1

        # Set middle bin index
        middle = low + (high - low) // 2

        if low == middle == high - 1:  # stop criterion
            return middle

        if value < array[middle]:  # array[low] <= value < array[middle]
            high = middle
        elif value >= array[middle]:  # array[middle] <= value < array[high]
            low = middle

    warnings.warn(RuntimeWarning("Operation took too long. Returning -1 (no match)."))
    return -1


def clean_feature_name(feature_name):
    """With the purpose to have a central feature cleaning, this function cleans
    feature names.

    Parameters
    ----------
    feature_name : string

    Returns
    -------
    cleaned_feature_name : string
    """
    return re.sub("[^a-z0-9]", "_", feature_name.lower())



from typing import TypeVar
from Amplo.GridSearch.BaseGridSearch import BaseGridSearch
from Amplo.GridSearch.HalvingGridSearch import HalvingGridSearch
from Amplo.GridSearch.OptunaGridSearch import OptunaGridSearch


GridSearchType = TypeVar('GridSearchType', BaseGridSearch, HalvingGridSearch,
                         OptunaGridSearch)



from abc import abstractmethod
import multiprocessing as mp
import re
from typing import Any, Dict, List, Tuple, Optional, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold
from sklearn.metrics import SCORERS
from sklearn.metrics._scorer import _BaseScorer  # noqa


__all__ = ["_GridSearch"]


class _GridSearch:
    def __init__(
        self,
        model,
        params=None,
        candidates=250,
        timeout=None,
        cv=KFold(n_splits=10),
        scoring="accuracy",
        verbose=0,
    ):
        """
        Abstract base class for grid search.

        Purposes:
            - Enforces to inheriting classes to implement all abstract methods.
            - Defines the hyperparameter search space as it's the same for all
              grid search methods.

        Parameters
        ----------
        model : Amplo.AutoML.Modeller.ModelType
            Model object to optimize.
        params : optional
            Parameters to optimize. Has no effect for `OptunaGridSearch`.
        candidates : int
            Limit the number of candidates to search.
        timeout : int
            Limit the time for optimization.
        cv : sklearn.model_selection.BaseCrossValidator
            Cross validation object.
        scoring : str or sklearn.metrics._scorer._BaseScorer
            A valid string for `sklearn.metrics.SCORERS`
        verbose : int
            Verbose logging.
        """
        # Input tests
        if hasattr(model, "is_fitted") and model.is_fitted:
            raise AssertionError("Model already fitted")
        if isinstance(scoring, str):
            self.scoring = SCORERS[scoring]
        elif not issubclass(type(scoring), _BaseScorer):
            raise ValueError(
                "Parameter `scoring` must originate from sklearn.metrics.make_scorer() "
                "or must be a valid string for sklearn.metrics.SCORERS."
            )

        # Set class attributes
        self.model = model
        self.params = params
        self.nTrials = candidates
        self.timeout = timeout
        self.cv = cv
        self.scoring = SCORERS[scoring] if isinstance(scoring, str) else scoring
        self.verbose = verbose

        self.x, self.y = None, None
        self.binary = True
        self.samples = None

        # Model specific settings
        if type(self.model).__name__ == "LinearRegression":
            self.nTrials = 1

    @property
    def _hyper_parameter_values(
        self,
    ) -> Dict[str, Tuple[str, List[Union[str, float]], Optional[int]]]:
        """Get model specific hyper parameter values, indicating predefined
        search areas to optimize.

        Notes
        -----

        Each item of the output dictionary consists of:
            - parameter name (str), and
            - parameter specifications (tuple)
                - parameter type (str)
                - parameter arguments (list)
                - number of distinct values (int, optional) [only for exhaustive grid
                  search]

        Parameter types include:
            - 'categorical': categorical values
            - 'int': discretized uniform value space
            - 'logint': discretized logarithmic uniform value space
            - 'uniform': uniform value space
            - 'loguniform': logarithmic uniform value space

        Parameter arguments are:
            - [categorical]: a list of all options
            - [int, logint, uniform, loguniform]: a tuple with min and max value

        **Special case (conditionals):**
            In some cases, one wants to grid-search certain parameters only
            if another parameter condition is present. Such conditions are
            specified via the dedicated key 'CONDITIONALS'.
        """

        # Extract model name & type
        model_name = type(self.model).__name__
        model_type = re.split(r"Regressor|Classifier", model_name)[0]
        # Determine whether it's classification or regression
        is_regression = bool(re.match(r".*(Regression|Regressor|SVR)", model_name))
        is_classification = bool(
            re.match(r".*(Classification|Classifier|SVC)", model_name)
        )
        assert (
            is_regression or is_classification
        ), "Could not determine mode (regression or classification)"

        # Define min-max-function
        def minimax(min_, value, max_):
            return max(min_, min(value, max_))

        # Find matching model and return its parameter values

        if model_name == "LinearRegression" or model_name == "LogisticRegression":
            return {}

        elif model_name == "Lasso" or "Ridge" in model_name:
            return dict(
                alpha=("uniform", [0, 10], 25),
            )

        elif model_name in ("SVR", "SVC"):
            return dict(
                gamma=("categorical", ["scale", "auto", 0.001, 0.01, 0.1, 0.5, 1]),
                C=("uniform", [0.001, 10], 25),
            )

        elif model_type == "KNeighbors":
            return dict(
                n_neighbors=("int", [5, minimax(5, self.samples // 10, 50)], 5),
                weights=("categorical", ["uniform", "distance"]),
                leaf_size=("int", [1, minimax(1, self.samples // 10, 100)], 5),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )

        elif model_type == "MLP":
            raise NotImplementedError("MLP is not supported")

        elif model_type == "SGD":
            params = dict(
                loss=(
                    "categorical",
                    [
                        "squared_loss",
                        "huber",
                        "epsilon_insensitive",
                        "squared_epsilon_insensitive",
                    ],
                ),
                penalty=("categorical", ["l2", "l1", "elasticnet"]),
                alpha=("uniform", [0, 10], 5),
                max_iter=("int", [250, 1000], 3),
            )
            if is_classification:
                params.update(
                    loss=(
                        "categorical",
                        ["hinge", "log", "modified_huber", "squared_hinge"],
                    )
                )
            return params

        elif model_type == "DecisionTree":
            params = dict(
                criterion=(
                    "categorical",
                    ["squared_error", "friedman_mse", "absolute_error", "poisson"],
                ),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 25)], 4),
            )
            if is_classification:
                params.update(criterion=("categorical", ["gini", "entropy"]))
            return params

        elif model_type == "AdaBoost":
            params = dict(
                n_estimators=("int", [25, 250], 5),
                loss=("categorical", ["linear", "square", "exponential"]),
                learning_rate=("loguniform", [0.001, 1], 10),
            )
            if is_classification:
                params.pop("loss", None)
            return params

        elif model_type == "Bagging":
            params = dict(
                n_estimators=("int", [10, 250], 4),
                max_samples=("uniform", [0.5, 1.0], 4),
                max_features=("uniform", [0.5, 1.0], 4),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )
            return params

        elif model_type == "CatBoost":
            params = dict(
                n_estimators=("int", [500, 2000], 5),
                verbose=("categorical", [0]),
                od_pval=("categorical", [1e-5]),
                loss_function=("categorical", ["MAE", "RMSE"]),
                learning_rate=("loguniform", [0.001, 0.5], 5),
                l2_leaf_reg=("uniform", [0, 10], 5),
                depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_data_in_leaf=("int", [1, minimax(1, self.samples // 10, 1000)], 5),
                grow_policy=(
                    "categorical",
                    ["SymmetricTree", "Depthwise", "Lossguide"],
                ),
            )
            if is_classification:
                params.update(
                    loss_function=(
                        "categorical",
                        ["Logloss" if self.binary else "MultiClass"],
                    )
                )
            return params

        elif model_type == "GradientBoosting":
            params = dict(
                loss=("categorical", ["ls", "lad", "huber"]),
                learning_rate=("loguniform", [0.001, 0.5], 10),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                n_estimators=("int", [100, 1000], 4),
                min_samples_leaf=(
                    "int",
                    [1, minimax(1, int(self.samples / 10), 1000)],
                    3,
                ),
                max_features=("uniform", [0.5, 1], 3),
                subsample=("uniform", [0.5, 1], 3),
            )
            if is_classification:
                params.update(loss=("categorical", ["deviance", "exponential"]))
            return params

        elif model_type == "HistGradientBoosting":
            params = dict(
                loss=("categorical", ["least_squares", "least_absolute_deviation"]),
                learning_rate=("loguniform", [0.001, 0.5], 10),
                max_iter=("int", [100, 250], 4),
                max_leaf_nodes=("int", [30, 150], 4),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_samples_leaf=(
                    "int",
                    [1, minimax(1, int(self.samples / 10), 1000)],
                    4,
                ),
                l2_regularization=("uniform", [0, 10], 5),
                max_bins=("int", [100, 255], 4),
                early_stopping=("categorical", [True]),
            )
            if is_classification:
                params.pop("loss", None)
            return params

        elif model_type == "RandomForest":
            params = dict(
                n_estimators=("int", [50, 1000], 5),
                criterion=("categorical", ["squared_error", "absolute_error"]),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 15)], 4),
                max_features=("categorical", ["auto", "sqrt"]),
                min_samples_split=("int", [2, 50], 4),
                min_samples_leaf=("int", [1, minimax(1, self.samples // 10, 1000)], 5),
                bootstrap=("categorical", [True, False]),
            )
            if is_classification:
                params.update(criterion=("categorical", ["gini", "entropy"]))
            return params

        elif model_type == "XGB":
            params = dict(
                objective=("categorical", ["reg:squarederror"]),
                eval_metric=("categorical", ["rmse"]),
                booster=("categorical", ["gbtree", "gblinear", "dart"]),
                alpha=("loguniform", [1e-8, 1], 10),
                learning_rate=("loguniform", [0.001, 0.5], 5),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )
            params["lambda"] = ("loguniform", [1e-8, 1], 10)
            if is_classification:
                params.update(
                    objective=("categorical", ["multi:softprob"]),
                    eval_metric=("categorical", ["logloss"]),
                )
            params["CONDITIONALS"] = dict(
                booster=[
                    (
                        "gbtree",
                        dict(
                            max_depth=(
                                "int",
                                [1, minimax(1, int(np.log2(self.samples)), 10)],
                                5,
                            ),
                            eta=("loguniform", [1e-8, 1], 5),
                            gamma=("loguniform", [1e-8, 1], 5),
                            grow_policy=("categorical", ["depthwise", "lossguide"]),
                        ),
                    ),
                    (
                        "dart",
                        dict(
                            max_depth=(
                                "int",
                                [1, minimax(1, int(np.log2(self.samples)), 10)],
                                5,
                            ),
                            eta=("loguniform", [1e-8, 1], 5),
                            gamma=("loguniform", [1e-8, 1], 5),
                            grow_policy=("categorical", ["depthwise", "lossguide"]),
                            sample_type=("categorical", ["uniform", "weighted"]),
                            normalize_type=("categorical", ["tree", "forest"]),
                            rate_drop=("loguniform", [1e-8, 1], 5),
                            skip_drop=("loguniform", [1e-8, 1], 5),
                        ),
                    ),
                ],
            )
            return params

        elif model_type == "LGBM":
            if is_regression:
                return dict(
                    num_leaves=("int", [10, 150], 5),
                    min_data_in_leaf=(
                        "int",
                        [1, minimax(1, self.samples // 10, 1000)],
                        0,
                    ),
                    min_sum_hessian_in_leaf=("uniform", [0.001, 0.5], 0),
                    colsample_bytree=("uniform", [0, 1], 5),
                    reg_alpha=("uniform", [0, 1], 5),
                    reg_lambda=("uniform", [0, 1], 5),
                    verbosity=("categorical", [-1]),
                    n_jobs=("categorical", [mp.cpu_count() - 1]),
                )
            else:  # is_classification
                return dict(
                    objective=(
                        "categorical",
                        ["binary" if self.binary else "multiclass"],
                    ),
                    metric=(
                        "categorical",
                        ["binary_error", "auc", "average_precision", "binary_logloss"]
                        if self.binary
                        else ["multi_error", "multi_logloss", "auc_mu"],
                    ),
                    boosting_type=("categorical", ["gbdt"]),
                    lambda_l1=("loguniform", [1e-8, 10], 4),
                    lambda_l2=("loguniform", [1e-8, 10], 4),
                    num_leaves=("int", [10, 5000], 4),
                    max_depth=("int", [5, 20], 4),
                    min_data_in_leaf=(
                        "int",
                        [1, minimax(1, self.samples // 10, 1000)],
                        0,
                    ),
                    min_gain_to_split=("uniform", [0, 5], 0),
                    feature_fraction=("uniform", [0.4, 1], 0),
                    bagging_fraction=("uniform", [0.4, 1], 0),
                    bagging_freq=("int", [1, 7], 0),
                    verbosity=("categorical", [-1]),
                    n_jobs=("categorical", [mp.cpu_count() - 1]),
                )

        # Raise error if no match was found
        raise NotImplementedError(
            "Hyper parameter tuning not implemented for {}".format(model_name)
        )

    def get_parameter_min_max(self) -> pd.DataFrame:
        """
        Get all min and max values from model-specific set of parameters.

        Omit categorical parameters as min and max values are ambiguous.

        Returns
        -------
        param_min_max : pd.DataFrame
            Min and max values.
        """

        # Get all model's parameters
        param_values = self._hyper_parameter_values
        # Pop conditionals and integrate all into `param_values`
        conditionals = param_values.pop("CONDITIONALS", {})
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                for name, value in additional_params.items():
                    param_values[name] = value

        # Filter for min and max in non-categorical parameters
        param_min_max = {}
        for p_name, value in param_values.items():
            p_type = value[0]
            p_args = value[1]
            if p_type in ("int", "logint", "uniform", "loguniform"):
                # Sanity check
                assert (
                    len(p_args) == 2
                ), "A {} should have a min and a max value".format(p_type)
                # Add item to dict
                add_item = {p_name: {"min": p_args[0], "max": p_args[1]}}
                param_min_max.update(add_item)

        # Combine all values to pd.DataFrame
        param_min_max = pd.DataFrame(param_min_max).T
        return param_min_max

    @abstractmethod
    def _get_hyper_params(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Get grid search specific distributions or samples.

        This function translates `self._hyper_parameter_values` to the expected
        format for the given grid search.

        Parameters
        ----------
        args : optional
            Grid search specific arguments.
        kwargs : optional
            Grid search specific keyword arguments.

        Returns
        -------
        Grid search parameters
        """
        pass

    @abstractmethod
    def fit(self, x, y) -> pd.DataFrame:
        """
        Run fit with model-specific set of parameters.

        Parameters
        ----------
        x : array
            Data features.
        y : array
            Data labels

        Returns
        -------
        results : pd.DataFrame
            Results of the grid search.
        """
        pass



import pytest
from pathlib import Path

from tests import rmtree


__all__ = ['TestAPI']


class TestAPI:
    sync_dir = Path('./test_dir')

    @pytest.fixture(autouse=True)
    def rmtree_sync_dir(self):
        rmtree(self.sync_dir)
        yield
        rmtree(self.sync_dir)



import pytest
from pathlib import Path

import numpy as np
import pandas as pd

from Amplo.AutoML import IntervalAnalyser
from Amplo.Utils.testing import make_interval_data
from tests import rmtree


def create_data_frames(n_samples, n_features):
    dim = (int(n_samples / 2), n_features)
    columns = [f'Feature_{i}' for i in range(n_features)]
    df1 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(100, 1, dim))))
    df2 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(-100, 1, dim))))
    return df1, df2


def create_test_folders(directory: Path, n_samples, n_features):
    # Make directories
    for sub_folder in ('Class_1', 'Class_2'):
        (directory / sub_folder).mkdir(exist_ok=True, parents=True)

    # Create and save dataframes
    for i in range(140):
        df1, df2 = create_data_frames(n_samples, n_features)
        df1.to_csv(directory / f'Class_1/Log_{i}.csv', index=False)
        df2.to_csv(directory / f'Class_2/Log_{i}.csv', index=False)


class TestIntervalAnalyser:

    @classmethod
    def setup_class(cls):
        cls.n_samples = 50
        cls.n_features = 25
        cls.ia_folder = Path('IA')
        rmtree(cls.ia_folder)
        create_test_folders(cls.ia_folder, cls.n_samples, cls.n_features)

    @classmethod
    def teardown_class(cls):
        rmtree(cls.ia_folder)

    def test_all(self):
        # Set up IntervalAnalyser
        ia = IntervalAnalyser(min_length=10)
        ia.fit_transform(self.ia_folder)

        # Attribute tests
        assert ia.n_folders == 2
        assert ia.n_files == 280

        # Functional tests
        for i in range(ia.n_files):
            dist = ia._distributions[i].values
            assert all(v < 0.8 for v in dist[:int(len(dist) / 2)]), 'Noise with high percentage of neighbors'
            assert all(v > 0.8 for v in dist[int(len(dist) / 2):]), 'Information with low percentage of neighbors'

        # Data tests
        data_no_noise = ia.data_without_noise
        data_full = ia.data_with_noise
        assert len(data_full) == ia.n_samples and len(data_no_noise) == ia.n_samples // 2, 'Incorrect number of samples'
        assert data_full.index.get_level_values(0).nunique() == ia.n_files, 'Files skipped'
        assert len(ia._features.columns) <= self.n_features, 'Incorrect number of features'

    def test_parse_data(self):
        # Test receiving dataframe and receiving directory
        for test in ('dataframe', 'single_dataframe', 'directory'):
            ia = IntervalAnalyser(target='labels')

            if test == 'dataframe':
                # Create dummy data
                features = make_interval_data(directory=None, target=ia.target, cat_choices=False)
                labels = features.pop(ia.target)
                # Set parse arguments
                parse_args = (features, labels)
            elif test == 'single_dataframe':
                # Create dummy data
                data = make_interval_data(directory=None, target=ia.target, cat_choices=False)
                # Set parse arguments
                parse_args = (data, )
            elif test == 'directory':
                # Create dummy data
                make_interval_data(directory=self.ia_folder, target=ia.target, cat_choices=False)
                # Set parse arguments
                parse_args = (self.ia_folder, None)
            else:
                raise NotImplementedError()

            # Read data
            _, _ = ia._parse_data(*parse_args)

            # Data checks
            assert all(isinstance(data, (pd.Series, pd.DataFrame))
                       for data in (ia._labels, ia._features, ia.orig_data)), 'Data should not be empty'
            assert len(ia._features.select_dtypes(include=['datetime64']).columns) == 0, \
                'Date-time columns should have been eliminated, internally'
            assert len(ia._features.select_dtypes(include=['float64']).columns) == 0, \
                'Dtype float64 columns should have been converted to float32'

            # Other checks
            assert not ia.is_fitted, 'IntervalAnalyser should not yet be fitted'
            assert ia._noise_indices is None, 'No noise indices should be set yet'

    # TODO test correlation warning
    # TODO test data quality warning
    # TODO test multi-index error
    # TODO test index misalignment error
    # TODO test feature/sample length error
    # TODO test target in features error
    # TODO test target != labels.name



import os
import json
import warnings
from typing import Union, Tuple
from pathlib import Path

import pandas as pd


__all__ = ['boolean_input', 'parse_json', 'read_pandas', 'merge_logs', 'get_log_metadata']


FILE_READERS = {'.csv': pd.read_csv, '.json': pd.read_json, '.xml': pd.read_xml,
                '.feather': pd.read_feather, '.parquet': pd.read_parquet,
                '.stata': pd.read_stata, '.pickle': pd.read_pickle}


def boolean_input(question: str) -> bool:
    x = input(question + ' [y / n]')
    if x.lower() == 'n' or x.lower() == 'no':
        return False
    elif x.lower() == 'y' or x.lower() == 'yes':
        return True
    else:
        print('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def parse_json(json_string: Union[str, dict]) -> Union[str, dict]:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(json_string
                              .replace("'", '"')
                              .replace("True", "true")
                              .replace("False", "false")
                              .replace("nan", "NaN")
                              .replace("None", "null"))
        except json.decoder.JSONDecodeError:
            print('[AutoML] Cannot validate, impassable JSON.')
            print(json_string)
            return json_string


def read_pandas(path: Union[str, Path]) -> pd.DataFrame:
    """
    Wrapper for various read functions

    Returns
    -------
    pd.DataFrame
    """
    file_extension = Path(path).suffix
    if file_extension not in FILE_READERS:
        raise NotImplementedError(f'File format {file_extension} not supported.')
    else:
        reader = FILE_READERS[file_extension]
        return reader(path, low_memory=False)


def merge_logs(path_to_folder, target='labels'):
    r"""
    Combine log files from given directory into a multi-indexed dataframe

    Notes
    -----
    Make sure that each protocol is located in a sub folder whose name represents the respective label.

    A directory structure example:
        |   ``path_to_folder``
        |   ``├─ Label_1``
        |   ``│   ├─ Log_1.*``
        |   ``│   └─ Log_2.*``
        |   ``├─ Label_2``
        |   ``│   └─ Log_3.*``
        |   ``└─ ...``

    Parameters
    ----------
    path_to_folder : str or Path
        Parent directory
    target : str
        Column name for target

    Returns
    -------
    data : pd.DataFrame
        All logs concatenated into one multi-indexed dataframe.
        Multi-index names are ``log`` and ``index``.
        Target column depicts the folder name.
    metadata : dict
        File metadata

    Warns
    --------
    EmptyDataError
        Whenever an empty file is found.
    NotImplementedError
        Whenever a file with an unknown format is found.
    """
    # Tests
    if not Path(path_to_folder).is_dir():
        raise ValueError(f'The provided path is no directory: {path_to_folder}')
    if not Path(path_to_folder).exists():
        raise FileNotFoundError(f'Directory does not exist: {path_to_folder}')
    if not isinstance(target, str) or target == '':
        raise ValueError('Target name must be a non-empty string.')

    # Result init
    data = []

    # Get file names
    metadata = get_log_metadata(path_to_folder)

    # Loop through file paths in metadata
    for file_id in metadata:
        # Read data
        try:
            datum = read_pandas(metadata[file_id]['full_path'])
        except pd.errors.EmptyDataError:
            warnings.warn(f"Empty file: {metadata[file_id]}")
            continue
        except NotImplementedError:
            warnings.warn(f"Unknown file format: {metadata[file_id]}")

        # Set labels
        datum[target] = metadata[file_id]['folder']

        # Set index
        datum.set_index(pd.MultiIndex.from_product([[file_id], datum.index.values], names=['log', 'index']),
                        inplace=True)

        # Add to list
        data.append(datum)

    if len(data) == 1:
        # Omit concatenation when only one item
        return data[0], metadata
    else:
        # Concatenate dataframes
        return pd.concat(data), metadata


def get_log_metadata(path_to_folder):
    """Get metadata of log files

    Parameters
    ----------
    path_to_folder

    Notes
    -----
    Make sure that each protocol is located in a sub folder whose name represents the respective label.

    A directory structure example:
        |   ``path_to_folder``
        |   ``├─ Label_1``
        |   ``│   ├─ Log_1.*``
        |   ``│   └─ Log_2.*``
        |   ``├─ Label_2``
        |   ``│   └─ Log_3.*``
        |   ``└─ ...``

    Returns
    -------
    metadata : dict
        Dictionary whose keys depict the file id (integer) and each value contains a dictionary with
        ``folder``, ``file``, ``full_path`` and ``last_modified`` key.
    """
    # Checks
    if not Path(path_to_folder).is_dir():
        raise ValueError(f'The provided path is no directory: {path_to_folder}')
    if not Path(path_to_folder).exists():
        raise FileNotFoundError(f'Directory does not exist: {path_to_folder}')

    # Init
    metadata = dict()
    file_id = 0

    # Loop through folders
    for folder in sorted(Path(path_to_folder).iterdir()):

        # Loop through files (ignore hidden files)
        for file in sorted(folder.glob('[!.]*.*')):

            # Check file
            if file.suffix not in FILE_READERS:
                warnings.warn(f'[AutoML] Skipped unsupported file format: {file}')
                continue
            elif file.stat().st_size == 0:
                warnings.warn(f'[AutoML] Skipped empty file: {file}')
                continue

            # Add to metadata
            metadata[file_id] = {
                'folder': str(folder.name),
                'file': str(file.name),
                'full_path': str(file.resolve()),
                'last_modified': os.path.getmtime(str(file)),
            }

            # Increment
            file_id += 1

    if file_id == 0:
        raise FileNotFoundError('Directory seems to be empty. Check whether you specified the correct path.')

    return metadata



import pytest

import pandas as pd
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression

from Amplo.AutoML import Modeller
from tests import rmtree


@pytest.fixture(scope='class', params=['classification', 'regression'])
def make_mode(request):
    mode = request.param
    if mode == 'classification':
        x, y = make_classification()
        objective = 'neg_log_loss'
    elif mode == 'regression':
        x, y = make_regression()
        objective = 'r2'
    else:
        raise ValueError('Invalid mode')
    request.cls.mode = mode
    request.cls.objective = objective
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.mark.usefixtures('make_mode')
class TestModelling:
    folder = 'tmp/'

    @pytest.fixture(autouse=True)
    def teardown(self):
        yield
        rmtree(self.folder)

    def test_modeller(self):
        mod = Modeller(mode=self.mode, objective=self.objective, folder=self.folder)
        mod.fit(self.x, self.y)
        # Tests
        assert isinstance(mod.results, pd.DataFrame), 'Results should be type pd.DataFrame'
        assert len(mod.results) != 0, 'Results empty'
        assert mod.results['mean_objective'].max() < 1, 'R2 needs to be smaller than 1'
        assert not mod.results['mean_objective'].isna().any(), "Mean Objective shouldn't contain NaN"
        assert not mod.results['std_objective'].isna().any(), "Std Objective shouldn't contain NaN"
        assert not mod.results['mean_time'].isna().any(), "Mean time shouldn't contain NaN"
        assert not mod.results['std_time'].isna().any(), "Std time shouldn't contain NaN"
        assert 'date' in mod.results.keys()
        assert 'model' in mod.results.keys()
        assert 'dataset' in mod.results.keys()
        assert 'params' in mod.results.keys()

    @pytest.mark.parametrize('n_samples', [100, 100_000])
    def test_return(self, n_samples):
        Modeller(mode=self.mode, samples=n_samples).return_models()



# Copyright by Amplo
from Amplo.Observation.base import ProductionWarning
from Amplo.Observation._data_observer import DataObserver
from Amplo.Observation._model_observer import ModelObserver
from Amplo.Observation.observer import ProductionObserver



import pytest

from sklearn.datasets import load_iris, make_regression

from Amplo.AutoML import DataExplorer
from tests import rmtree


@pytest.fixture(scope='class', params=['classification', 'regression'])
def make_mode(request):
    mode = request.param
    if mode == 'classification':
        x, y = load_iris(return_X_y=True, as_frame=True)
    elif mode == 'regression':
        # Interesting enough, the regression test for `test_explorer` ends up in a deadlock,
        #  specifically in the `DataExplorer.shap()` method. It's even more interesting that
        #  the california housing dataset works fine when using `unittest`
        # x, y = fetch_california_housing(return_X_y=True, as_frame=True)
        x, y = make_regression()
    else:
        raise NotImplementedError('Invalid mode')
    request.cls.x = x
    request.cls.y = y
    request.cls.mode = mode
    yield


@pytest.mark.usefixtures('make_mode')
class TestDataExploring:

    def test_explorer(self):
        tmp_folder = 'tmp/'
        eda = DataExplorer(self.x, y=self.y, mode=self.mode, folder=tmp_folder)
        eda.run()
        rmtree(tmp_folder)


# class TestDataExploring(unittest.TestCase):
#
#     @classmethod
#     def setUpClass(cls):
#         cls.class_x, cls.class_y = load_iris(return_X_y=True, as_frame=True)
#         cls.reg_x, cls.reg_y = fetch_california_housing(return_X_y=True)
#
#     def test_regression(self):
#         eda = DataExplorer(self.reg_x, y=self.reg_y, mode='regression', folder='tmp/')
#         eda.run()
#         rmtree('tmp/')
#
#     def test_classification(self):
#         eda = DataExplorer(self.class_x, y=self.class_y, mode='classification', folder='tmp/')
#         eda.run()
#         rmtree('tmp/')
